#ifndef HARKIO_HARKCONFIG_TSPLISTFILE_HPP_
#define HARKIO_HARKCONFIG_TSPLISTFILE_HPP_

#include <string>
#include <vector>
#include <glibmm/ustring.h>
#include <harkio2/Version.hpp>
#include <harkio2/defineitem.hpp>
#include <harkio2/FileType.hpp>
#include <harkio2/TypedFilename.hpp>
#include <harkio2/harkconfig/HarkConfigFile.hpp>
#include <harkio2/harkconfig/Source.hpp>
#include <harkio2/harkconfig/HarkConfigFileWriter.hpp>

/** @file TSPListFile.hpp
 * @brief TSPListFile class and subclasses definitions.
 */

namespace harkio2
{
namespace harkconfig
{

/** @brief Abstract TSP list file.
 * This class represents a file which contains a list of TSP.
 * The file can be any version.
 */
class TSPListFile : public HarkConfigFile
{
public:
    /** @brief Returns the file type identifier.
     * @return The file type identifier which this class represents.
     */
    virtual std::string GetId() const { return "TSPListFile"; }
};

/** @brief TSP list file version 1.0.
 */
class TSPListFile1_0 : public TSPListFile
{
public:
    /** @brief Returns the file version.
     * @return The version which this class represents.
     */
    virtual harkio2::Version GetVersion() const { return Version(1, 0); }

    DEFINE_ITEM(synchronousAverage, int, SynchronousAverage, int)
    DEFINE_ITEM(originalImpulseFltFile, std::string, OriginalImpulseFltFile, const std::string&)
    DEFINE_ITEM(tspOffset, int, TSPOffset, int)
    DEFINE_ITEM(signalMax, int, SignalMax, int)
    DEFINE_ITEM(channelList, std::vector<int>, ChannelList, const std::vector<int>&)
    DEFINE_ITEM(sources, std::vector<harkio2::harkconfig::Source1>, Sources, const std::vector<harkio2::harkconfig::Source1>&)

    /// @brief Implementation of Writable interface.
    void Write(HarkConfigFileWriter& writer) const throw (harkio2::InvalidHarkFile)
    {
        writer.Write(*this);
    }

};

/** @brief TSP list file version 2.0.
 */
class TSPListFile2_0 : public TSPListFile
{
public:
    /** @brief Returns the file version.
     * @return The version which this class represents.
     */
    virtual harkio2::Version GetVersion() const { return Version(2, 0); }

    DEFINE_ITEM(synchronousAverage, unsigned int, SynchronousAverage, unsigned int)
    DEFINE_ITEM(originalImpulseFile, harkio2::TypedFilename, OriginalImpulseFile, const harkio2::TypedFilename&)
    DEFINE_ITEM(tspOffset, unsigned int, TSPOffset, unsigned int)
    DEFINE_ITEM(tspLength, unsigned int, TSPLength, unsigned int)
    DEFINE_ITEM(signalMax, unsigned int, SignalMax, unsigned int)
    DEFINE_ITEM(numMic, unsigned int, NumMic, unsigned int)
    DEFINE_ITEM(sourceFileType, harkio2::FileType::Kind, SourceFileType, harkio2::FileType::Kind)
    DEFINE_ITEM(comment, Glib::ustring, Comment, const Glib::ustring&)
    DEFINE_ITEM(sources, std::vector<harkio2::harkconfig::Source2>, Sources, const std::vector<harkio2::harkconfig::Source2>&)

    /// @brief Implementation of Writable interface.
    void Write(HarkConfigFileWriter& writer) const throw (harkio2::InvalidHarkFile)
    {
        writer.Write(*this);
    }

};

}
}

#endif // HARKIO_HARKCONFIG_TSPLISTFILE_HPP_

