#include <vector>
#include <string>
#include <cstring>
#include <stdint.h>
#include <boost/shared_ptr.hpp>
#include <boost/format.hpp>
#include <harkio2/AnalyzeError.hpp>
#include <harkio2/Endian.hpp>
#include <harkio2/Version.hpp>
#include <harkio2/Vector3.hpp>
#include <harkio2/hark/HarkFileBinaryReader.hpp>
#include <harkio2/hark/DataSet.hpp>
#include <harkio2/hark/HarkFile.hpp>
#include <harkio2/hark/hgtf/Fmt.hpp>
#include <harkio2/hark/hgtf/Spec.hpp>
#include <harkio2/hark/hgtf/M2PG.hpp>
#include <harkio2/hark/hgtf/GTF.hpp>
#include <harkio2/hark/hgtf/SM.hpp>

/** @file Helper.cpp
 * @brief implementation of HarkFileBinaryReader::Helper.
 */

namespace harkio2
{
namespace hark
{

class HarkFileBinaryReader::Helper
{
    HarkFileBinaryReader& parent;
public:
    /** @brief Constructs a helper for HarkFileBinaryReader.
     * @param parent A reference to a parent instance.
     */
    Helper(HarkFileBinaryReader& parent)
        : parent(parent)
    { }

    struct ChunkHead
    {
        char id[4];
        uint32_t size;
    };

    /** @brief Reads sizeof(T) bytes from parent's input stream.
     * This function takes account of endians.
     * @param[out] result The read value.
     * @throw harkio2::AnalyzeError Couldn't read sizeof(T) bytes.
     */
    template <typename T>
    void Read(T& result) throw (harkio2::AnalyzeError)
    {
        parent.in.read(reinterpret_cast<char*>(&result), sizeof(T));
        result = harkio2::Endian::LittleToHost(result);
        if (parent.in.gcount() != sizeof(T))
        {
            throw harkio2::AnalyzeError((boost::format("HarkFileBinaryReader::Helper::Read: cannot read %d bytes") % sizeof(T)).str());
        }
    }

    /** @brief Reads n bytes from parent's input stream.
     * This function doesn't convert byte order.
     * @param[out] buf The buffer for read data.
     * @param[in] n The number of bytes to read.
     * @throw harkio2::AnalyzeError Couldn't read n bytes.
     */
    void Read(char* buf, size_t n) throw (harkio2::AnalyzeError)
    {
        parent.in.read(buf, n);
        if (static_cast<size_t>(parent.in.gcount()) != n)
        {
            throw harkio2::AnalyzeError((boost::format("HarkFileBinaryReader::Helper::Read: cannot read %d bytes") % n).str());
        }
    }

    /** @brief Reads n bytes from parent's input stream and writes NUL to the end of buffer.
     * This function doesn't convert byte order.
     * Writes '\0' to buf[n].
     * @param[out] buf The buffer for read data.
     * @param[in] n The number of bytes to read.
     * @throw harkio2::AnalyzeError Couldn't read n bytes.
     */
    void ReadCString(char* buf, size_t n) throw (harkio2::AnalyzeError)
    {
        Read(buf, n);
        buf[n] = '\0';
    }

    /** @brief Reads n bytes from parent's input stream and seeks to previous position.
     * This function doesn't convert byte order.
     * @param[out] buf The buffer for peeked data.
     * @param[in] n The number of bytes to peek.
     * @throw harkio2::AnalyzeError Couldn't peak n bytes.
     */
    void Peek(char* buf, size_t n) throw (harkio2::AnalyzeError)
    {
        std::streampos pos = parent.in.tellg();
        Read(buf, n);
        parent.in.seekg(pos, std::ios::beg);
    }

    /** @brief Returns a chunk header read from input stream.
     * @return A read chunk header.
     * @throw harkio2::AnalyzeError Couldn't read successfully.
     */
    ChunkHead ReadHead() throw (harkio2::AnalyzeError)
    {
        ChunkHead result;
        Read(result.id, 4);
        Read(result.size);
        return result;
    }

    /** @brief Returns a chunk version read from input stream.
     * @return A read chunk version.
     * @throw harkio2::AnalyzeError Couldn't read successfully.
     */
    harkio2::Version ReadVersion() throw (harkio2::AnalyzeError)
    {
        uint16_t major_, minor_;
        Read(major_);
        Read(minor_);
        return harkio2::Version(major_, minor_);
    }

    /** @brief Returns a chunk header read from input stream and checks the head data.
     * @param chunkName a expected chunk name.
     * @param minSize a expected minimum chunk size.
     * @return A read chunk header.
     * @throw harkio2::AnalyzeError Couldn't read successfully.
     */
    ChunkHead ReadCheckHead(const std::string& chunkName, uint32_t minSize) throw (harkio2::AnalyzeError)
    {
        const char* funcName = "HarkFileBinaryReader::Helper::ReadCheckHead";
        ChunkHead head = ReadHead();
        if (strncmp(head.id, chunkName.c_str(), 4) != 0)
        {
            throw harkio2::AnalyzeError((boost::format("%s: this is not %s chunk") % funcName % chunkName).str());
        }
        else if (head.size < minSize)
        {
            throw harkio2::AnalyzeError((boost::format("%s: invalid size property") % funcName).str());
        }
        return head;
    }

    /** @brief Returns a chunk version read from input stream and checks the version.
     * @param expected a expected chunk version.
     * @return A read chunk version.
     * @throw harkio2::AnalyzeError Couldn't read successfully.
     */
    harkio2::Version ReadCheckVersion(const harkio2::Version& expected) throw (harkio2::AnalyzeError)
    {
        const char* funcName = "HarkFileBinaryReader::Helper::ReadCheckVersion";
        harkio2::Version v = ReadVersion();
        if (v != expected)
        {
            throw harkio2::AnalyzeError((boost::format("%s: version is not %s") % funcName % expected.ToString()).str());
        }
        return v;
    }

    /** @brief Returns a Fmt read from input stream.
     * @return A read Fmt data structure.
     * @throw harkio2::AnalyzeError Couldn't read successfully.
     */
    boost::shared_ptr<harkio2::hark::hgtf::Fmt> ReadFmt() throw (harkio2::AnalyzeError)
    {
        const char* funcName = "HarkFileBinaryReader::Helper::ReadFmt";

        std::streampos pos = parent.in.tellg();
        ChunkHead head = ReadHead();
        harkio2::Version v = ReadVersion();
        parent.in.seekg(pos, std::ios::beg);

        if (strncmp(head.id, "fmt ", 4) != 0)
        {
            throw harkio2::AnalyzeError((boost::format("%s: this is not fmt chunk") % funcName).str());
        }

        if (v == harkio2::Version(1, 0))
        {
            return ReadFmt1_0();
        }
        else
        {
            throw harkio2::AnalyzeError((boost::format("%s: unknown version: %s") % funcName % v.ToString()).str());
        }
    }

    /** @brief Returns a Fmt1_0 read from input stream.
     * @return A read Fmt1_0 data structure.
     * @throw harkio2::AnalyzeError Couldn't read successfully.
     */
    boost::shared_ptr<harkio2::hark::hgtf::Fmt1_0> ReadFmt1_0() throw (harkio2::AnalyzeError)
    {
        const char* funcName = "HarkFileBinaryReader::Helper::ReadFmt1_0";

        ChunkHead head = ReadCheckHead("fmt ", 24);

        ReadCheckVersion(harkio2::Version(1, 0));

        uint32_t formatId, nMic, nFFT, nIR, fs;
        Read(formatId);
        Read(nMic);
        Read(nFFT);
        Read(nIR);
        Read(fs);

        if (head.size != (24 + 12 * nMic))
        {
            throw harkio2::AnalyzeError((boost::format("%s: invalid size property") % funcName).str());
        }
        else if (formatId <= 0 || 4 <= formatId)
        {
            throw harkio2::AnalyzeError((boost::format("%s: invalid format identifier") % funcName).str());
        }

        std::vector<harkio2::Vector3<float> > mics;
        mics.reserve(nMic);

        float x, y, z;
        for (size_t i = 0; i < nMic; ++i)
        {
            Read(x);
            Read(y);
            Read(z);
            mics.push_back(harkio2::Vector3<float>(x, y, z));
        }

        boost::shared_ptr<harkio2::hark::hgtf::Fmt1_0> result(new harkio2::hark::hgtf::Fmt1_0());
        result->SetFormatId(formatId);
        result->SetNMic(nMic);
        result->SetNFFT(nFFT);
        result->SetNImpulseResponse(nIR);
        result->SetFS(fs);
        result->SetMicPositions(mics);

        return result;
    }

    /** @brief Returns a Spec read from input stream.
     * @return A read Spec data structure.
     * @throw harkio2::AnalyzeError Couldn't read successfully.
     */
    boost::shared_ptr<harkio2::hark::hgtf::Spec> ReadSpec() throw (harkio2::AnalyzeError)
    {
        const char* funcName = "HarkFileBinaryReader::Helper::ReadSpec";

        std::streampos pos = parent.in.tellg();
        ChunkHead head = ReadHead();
        harkio2::Version v = ReadVersion();
        parent.in.seekg(pos, std::ios::beg);

        if (strncmp(head.id, "SPEC", 4) != 0)
        {
            throw harkio2::AnalyzeError((boost::format("%s: this is not SPEC chunk") % funcName).str());
        }

        if (v == harkio2::Version(1, 0))
        {
            return ReadSpec1_0();
        }
        else
        {
            throw harkio2::AnalyzeError((boost::format("%s: unknown version: %s") % funcName % v.ToString()).str());
        }
    }

    /** @brief Returns a Spec1_0 read from input stream.
     * @return A read Spec1_0 data structure.
     * @throw harkio2::AnalyzeError Couldn't read successfully.
     */
    boost::shared_ptr<harkio2::hark::hgtf::Spec1_0> ReadSpec1_0() throw (harkio2::AnalyzeError)
    {
        const char* funcName = "HarkFileBinaryReader::Helper::ReadSpec1_0";

        ChunkHead head = ReadCheckHead("SPEC", 1551);
        if (head.size != 1551)
        {
            throw harkio2::AnalyzeError((boost::format("%s: invalid size property") % funcName).str());
        }

        ReadCheckVersion(harkio2::Version(1, 0));

        boost::shared_ptr<harkio2::hark::hgtf::Spec1_0> result(new harkio2::hark::hgtf::Spec1_0());
        char buf[1025];
        ReadCString(buf, 11);
        result->SetDate(buf);
        ReadCString(buf, 128);
        result->SetLocation(buf);
        ReadCString(buf, 128);
        result->SetRobot(buf);
        ReadCString(buf, 128);
        result->SetMicArrayType(buf);
        ReadCString(buf, 128);
        result->SetCreator(buf);
        ReadCString(buf, 1024);
        result->SetMemo(buf);

        return result;
    }

    /** @brief Returns a M2PG read from input stream.
     * @param nMic The number of microphones specified by Fmt.
     * @param nFFT The number of FFT points specified by Fmt.
     * @return A read M2PG data structure.
     * @throw harkio2::AnalyzeError Couldn't read successfully.
     */
    boost::shared_ptr<harkio2::hark::hgtf::M2PG> ReadM2PG(unsigned int nMic, unsigned int nFFT) throw (harkio2::AnalyzeError)
    {
        const char* funcName = "HarkFileBinaryReader::Helper::ReadM2PG";

        std::streampos pos = parent.in.tellg();
        ChunkHead head = ReadHead();
        harkio2::Version v = ReadVersion();
        parent.in.seekg(pos, std::ios::beg);

        if (strncmp(head.id, "M2PG", 4) != 0)
        {
            throw harkio2::AnalyzeError((boost::format("%s: this is not M2PG chunk") % funcName).str());
        }

        if (v == harkio2::Version(1, 0))
        {
            return ReadM2PG1_0(nMic, nFFT);
        }
        else
        {
            throw harkio2::AnalyzeError((boost::format("%s: unknown version: %s") % funcName % v.ToString()).str());
        }
    }

    /** @brief Returns a M2PG1_0 read from input stream.
     * @param nMic The number of microphones specified by Fmt.
     * @param nFFT The number of FFT points specified by Fmt.
     * @return A read M2PG1_0 data structure.
     * @throw harkio2::AnalyzeError Couldn't read successfully.
     */
    boost::shared_ptr<harkio2::hark::hgtf::M2PG1_0> ReadM2PG1_0(unsigned int nMic, unsigned int nFFT) throw (harkio2::AnalyzeError)
    {
        const char* funcName = "HarkFileBinaryReader::Helper::ReadM2PG1_0";

        ChunkHead head = ReadCheckHead("M2PG", 12);
        ReadCheckVersion(harkio2::Version(1, 0));

        uint32_t nd, nr;

        Read(nd);
        Read(nr);

        if (head.size != 12 + 12*nd*nr + 8*nd*nr*nMic*(nFFT/2 + 1))
        {
            throw harkio2::AnalyzeError((boost::format("%s: invalid size property") % funcName).str());
        }

        std::vector<std::vector<harkio2::Vector3<float> > > positions(
                nd, std::vector<harkio2::Vector3<float> >(nr));
        std::vector<std::vector<std::vector<harkio2::hark::hgtf::M2PG1_0::TFType> > >
            tfs(nd, std::vector<std::vector<harkio2::hark::hgtf::M2PG1_0::TFType> >(
                        nr, std::vector<harkio2::hark::hgtf::M2PG1_0::TFType>(
                            nMic, harkio2::hark::hgtf::M2PG1_0::TFType(nFFT/2+1))));

        float x, y, z;
        for (size_t d = 0; d < nd; ++d)
        {
            for (size_t r = 0; r < nr; ++r)
            {
                Read(x);
                Read(y);
                Read(z);
                positions[d][r] = harkio2::Vector3<float>(x, y, z);
            }
        }

        float real, imag;
        for (size_t d = 0; d < nd; ++d)
        {
            for (size_t r = 0; r < nr; ++r)
            {
                for (size_t m = 0; m < nMic; ++m)
                {
                    for (size_t f = 0; f < nFFT/2 + 1; ++f)
                    {
                        Read(real);
                        Read(imag);
                        tfs[d][r][m][f] = std::complex<float>(real, imag);
                    }
                }
            }
        }

        boost::shared_ptr<harkio2::hark::hgtf::M2PG1_0> result(new harkio2::hark::hgtf::M2PG1_0());
        result->SetND(nd);
        result->SetNR(nr);
        result->SetSourcePositions(positions);
        result->SetTFs(tfs);

        return result;
    }

    /** @brief Returns a GTF read from input stream.
     * @param nMic The number of microphones specified by Fmt.
     * @return A read GTF data structure.
     * @throw harkio2::AnalyzeError Couldn't read successfully.
     */
    boost::shared_ptr<harkio2::hark::hgtf::GTF> ReadGTF(unsigned int nMic) throw (harkio2::AnalyzeError)
    {
        const char* funcName = "HarkFileBinaryReader::Helper::ReadGTF";

        std::streampos pos = parent.in.tellg();
        ChunkHead head = ReadHead();
        harkio2::Version v = ReadVersion();
        parent.in.seekg(pos, std::ios::beg);

        if (strncmp(head.id, "GTF ", 4) != 0)
        {
            throw harkio2::AnalyzeError((boost::format("%s: this is not GTF chunk") % funcName).str());
        }

        if (v == harkio2::Version(1, 0))
        {
            return ReadGTF1_0(nMic);
        }
        else
        {
            throw harkio2::AnalyzeError((boost::format("%s: unknown version: %s") % funcName % v.ToString()).str());
        }
    }

    /** @brief Returns a GTF1_0 read from input stream.
     * @param nMic The number of microphones specified by Fmt.
     * @return A read GTF1_0 data structure.
     * @throw harkio2::AnalyzeError Couldn't read successfully.
     */
    boost::shared_ptr<harkio2::hark::hgtf::GTF1_0> ReadGTF1_0(unsigned int nMic) throw (harkio2::AnalyzeError)
    {
        const char* funcName = "HarkFileBinaryReader::Helper::ReadGTF1_0";

        ChunkHead head = ReadCheckHead("GTF ", 12);
        ReadCheckVersion(harkio2::Version(1, 0));

        uint32_t nSource;

        Read(nSource);

        std::vector<harkio2::Vector3<float> > positions(nSource);

        float x, y, z;
        for (size_t i = 0; i < nSource; ++i)
        {
            Read(x);
            Read(y);
            Read(z);
            positions[i] = harkio2::Vector3<float>(x, y, z);
        }

        uint32_t nFIndex;

        Read(nFIndex);

        if (head.size != 8 + 12*nSource + 4 + 4*nFIndex + 8*nSource*nMic*nFIndex)
        {
            throw harkio2::AnalyzeError((boost::format("%s: invalid size property") % funcName).str());
        }

        std::vector<unsigned int> fbinmap(nFIndex);
        uint32_t tfIndex;

        for (size_t i = 0; i < nFIndex; ++i)
        {
            Read(tfIndex);
            fbinmap[i] = static_cast<unsigned int>(tfIndex);
        }

        std::vector<std::vector<harkio2::hark::hgtf::GTF1_0::TFType> > tfs(
                nSource, std::vector<harkio2::hark::hgtf::GTF1_0::TFType>(
                    nMic, harkio2::hark::hgtf::GTF1_0::TFType(nFIndex)));
        float real, imag;
        for (size_t s = 0; s < nSource; ++s)
        {
            for (size_t m = 0; m < nMic; ++m)
            {
                for (size_t f = 0; f < nFIndex; ++f)
                {
                    Read(real);
                    Read(imag);
                    tfs[s][m][f] = std::complex<float>(real, imag);
                }
            }
        }

        boost::shared_ptr<harkio2::hark::hgtf::GTF1_0> result(new harkio2::hark::hgtf::GTF1_0());
        result->SetNSource(nSource);
        result->SetSourcePositions(positions);
        result->SetNFIndex(nFIndex);
        result->SetFIndexToBinIndexTable(fbinmap);
        result->SetTFs(tfs);

        return result;
    }

    /** @brief Returns a SM read from input stream.
     * @param nMic The number of microphones specified by SM.
     * @return A read SM data structure.
     * @throw harkio2::AnalyzeError Couldn't read successfully.
     */
    boost::shared_ptr<harkio2::hark::hgtf::SM> ReadSM(unsigned int nMic) throw (harkio2::AnalyzeError)
    {
        const char* funcName = "HarkFileBinaryReader::Helper::ReadSM";

        std::streampos pos = parent.in.tellg();
        ChunkHead head = ReadHead();
        harkio2::Version v = ReadVersion();
        parent.in.seekg(pos, std::ios::beg);

        if (strncmp(head.id, "SM  ", 4) != 0)
        {
            throw harkio2::AnalyzeError((boost::format("%s: this is not SM chunk") % funcName).str());
        }

        if (v == harkio2::Version(1, 0))
        {
            return ReadSM1_0(nMic);
        }
        else
        {
            throw harkio2::AnalyzeError((boost::format("%s: unknown version: %s") % funcName % v.ToString()).str());
        }
    }

    /** @brief Returns a SM1_0 read from input stream.
     * @param nMic The number of microphones specified by Fmt.
     * @return A read SM1_0 data structure.
     * @throw harkio2::AnalyzeError Couldn't read successfully.
     */
    boost::shared_ptr<harkio2::hark::hgtf::SM1_0> ReadSM1_0(unsigned int nMic) throw (harkio2::AnalyzeError)
    {
        const char* funcName = "HarkFileBinaryReader::Helper::ReadSM1_0";

        ChunkHead head = ReadCheckHead("SM  ", 12);
        ReadCheckVersion(harkio2::Version(1, 0));

        uint32_t nSource;

        Read(nSource);

        std::vector<harkio2::Vector3<float> > positions(nSource);

        float x, y, z;
        for (size_t i = 0; i < nSource; ++i)
        {
            Read(x);
            Read(y);
            Read(z);
            positions[i] = harkio2::Vector3<float>(x, y, z);
        }

        uint32_t nFIndex;

        Read(nFIndex);

        if (head.size != 8 + 12*nSource + 4 + 4*nFIndex + 8*nSource*nMic*nFIndex)
        {
            throw harkio2::AnalyzeError((boost::format("%s: invalid size property") % funcName).str());
        }

        std::vector<unsigned int> fbinmap(nFIndex);
        uint32_t tfIndex;

        for (size_t i = 0; i < nFIndex; ++i)
        {
            Read(tfIndex);
            fbinmap[i] = static_cast<unsigned int>(tfIndex);
        }

        std::vector<harkio2::hark::hgtf::SM1_0::SMType> sms(
                nSource, harkio2::hark::hgtf::SM1_0::SMType(
                    nMic, std::vector<std::complex<float> >(nFIndex)));
        float real, imag;
        for (size_t s = 0; s < nSource; ++s)
        {
            for (size_t m = 0; m < nMic; ++m)
            {
                for (size_t f = 0; f < nFIndex; ++f)
                {
                    Read(real);
                    Read(imag);
                    sms[s][m][f] = std::complex<float>(real, imag);
                }
            }
        }

        boost::shared_ptr<harkio2::hark::hgtf::SM1_0> result(new harkio2::hark::hgtf::SM1_0());
        result->SetNSource(nSource);
        result->SetSourcePositions(positions);
        result->SetNFIndex(nFIndex);
        result->SetFIndexToBinIndexTable(fbinmap);
        result->SetSMs(sms);

        return result;
    }

    boost::shared_ptr<harkio2::hark::HarkFile> ReadHGTF()
    {
        const char* funcName = "HarkFileBinaryReader::Helper::ReadHGTF";

        char tag[4];
        Read(tag, 4);
        if (strncmp(tag, "HGTF", 4) != 0)
        {
            throw harkio2::AnalyzeError((boost::format("%s: this is not HGTF file") % funcName).str());
        }

        boost::shared_ptr<harkio2::hark::HarkFile> result(new harkio2::hark::HarkFile());
        bool foundFmt = false; // Fmt's information are needed for reading M2PG, GTF, SM
        unsigned int nMic = 0, nFFT = 0;

        // can't use Peek() because Peek() throw an exception if it can't read successfully.
        parent.in.read(tag, 4);
        while (parent.in.gcount() == 4)
        {
            parent.in.seekg(-4, std::ios::cur);

            if (strncmp(tag, "fmt ", 4) == 0)
            {
                boost::shared_ptr<harkio2::hark::hgtf::Fmt> fmt = ReadFmt();
                result->AddDataSet(fmt);
                foundFmt = true;

                // get nMic and nFFT
                if (fmt->GetVersion() == harkio2::Version(1, 0))
                {
                    boost::shared_ptr<harkio2::hark::hgtf::Fmt1_0> fmt10 =
                        boost::dynamic_pointer_cast<harkio2::hark::hgtf::Fmt1_0>(fmt);
                    nMic = fmt10->GetNMic();
                    nFFT = fmt10->GetNFFT();
                }
            }
            else if (strncmp(tag, "SPEC", 4) == 0)
            {
                result->AddDataSet(ReadSpec());
            }
            else if (strncmp(tag, "M2PG", 4) == 0)
            {

                if (foundFmt)
                {
                    result->AddDataSet(ReadM2PG(nMic, nFFT));
                }
                else
                {
                    throw harkio2::AnalyzeError((boost::format("%s: fmt chunk must exist before M2PG chunk") % funcName).str());
                }
            }
            else if (strncmp(tag, "GTF ", 4) == 0)
            {

                if (foundFmt)
                {
                    result->AddDataSet(ReadGTF(nMic));
                }
                else
                {
                    throw harkio2::AnalyzeError((boost::format("%s: fmt chunk must exist before GTF chunk") % funcName).str());
                }
            }
            else if (strncmp(tag, "SM  ", 4) == 0)
            {

                if (foundFmt)
                {
                    result->AddDataSet(ReadSM(nMic));
                }
                else
                {
                    throw harkio2::AnalyzeError((boost::format("%s: fmt chunk must exist before SM chunk") % funcName).str());
                }
            }

            parent.in.read(tag, 4);
        }

        return result;
    }

    boost::shared_ptr<harkio2::hark::HarkFile> ReadHGFB()
    {
        const char* funcName = "HarkFileBinaryReader::Helper::ReadHGFB";
        return boost::shared_ptr<harkio2::hark::HarkFile>();
    }
};

}
}

