#define BOOST_TEST_MODULE "testcase for HarkFileBinaryReader::Helper class"
#define BOOST_TEST_DYN_LINK
#include "common.hpp"

#include <boost/test/unit_test.hpp>

#define HARKIO_UNITTEST

#include <sstream>
#include <harkio2/hark/HarkFileReader.hpp>

#include "../src/hark/HarkFileBinaryReader/Helper.cpp"

template <typename T>
void Write(std::stringstream& ss, T val)
{
    ss.write(reinterpret_cast<char*>(&val), sizeof(T));
}

void WriteString(std::stringstream& ss, const std::string& s)
{
    ss.write(s.c_str(), s.length());
}

void WriteString(std::stringstream& ss, const Glib::ustring& s)
{
    ss.write(s.c_str(), s.bytes());
}

void MakeTestData(std::stringstream& ss, const harkio2::hark::hgtf::Fmt1_0& fmt)
{
    ss.write("fmt ", 4);
    Write(ss, static_cast<uint32_t>(24 + 12 * fmt.GetNMic()));
    Write(ss, static_cast<uint16_t>(1));
    Write(ss, static_cast<uint16_t>(0));
    Write(ss, static_cast<uint32_t>(fmt.GetFormatId()));
    Write(ss, static_cast<uint32_t>(fmt.GetNMic()));
    Write(ss, static_cast<uint32_t>(fmt.GetNFFT()));
    Write(ss, static_cast<uint32_t>(fmt.GetNImpulseResponse()));
    Write(ss, static_cast<uint32_t>(fmt.GetFS()));

    for (size_t i = 0; i < fmt.GetNMic(); ++i)
    {
        const harkio2::Vector3<float>& pos = fmt.GetMicPositions().at(i);
        Write(ss, pos.GetX());
        Write(ss, pos.GetY());
        Write(ss, pos.GetZ());
    }

    ss.seekp(0, std::ios::beg);
}

void MakeTestData(std::stringstream& ss, const harkio2::hark::hgtf::Spec1_0& spec)
{
    std::streampos base = ss.tellp();

    ss.write("SPEC", 4);
    Write(ss, static_cast<uint32_t>(1551));
    Write(ss, static_cast<uint16_t>(1));
    Write(ss, static_cast<uint16_t>(0));

    for (size_t i = 0; i < 1551 - 4; ++i)
    {
        ss.put('\0');
    }

    ss.seekp(base + std::streamoff(12));
    WriteString(ss, spec.GetDate());
    ss.seekp(base + std::streamoff(23));
    WriteString(ss, spec.GetLocation());
    ss.seekp(base + std::streamoff(23+128*1));
    WriteString(ss, spec.GetRobot());
    ss.seekp(base + std::streamoff(23+128*2));
    WriteString(ss, spec.GetMicArrayType());
    ss.seekp(base + std::streamoff(23+128*3));
    WriteString(ss, spec.GetCreator());
    ss.seekp(base + std::streamoff(23+128*4));
    WriteString(ss, spec.GetMemo());

    ss.seekp(0, std::ios::beg);
}

void MakeTestData(std::stringstream& ss, const harkio2::hark::hgtf::M2PG1_0& m2pg, uint32_t nMic, uint32_t nFFT)
{
    ss.write("M2PG", 4);
    uint32_t nd = m2pg.GetND();
    uint32_t nr = m2pg.GetNR();
    Write(ss, static_cast<uint32_t>(12 + 12*nd*nr + 8*nd*nr*nMic*(nFFT/2 + 1)));
    Write(ss, static_cast<uint16_t>(1));
    Write(ss, static_cast<uint16_t>(0));

    Write(ss, nd);
    Write(ss, nr);

    for (size_t d = 0; d < nd; ++d)
    {
        for (size_t r = 0; r < nr; ++r)
        {
            const harkio2::Vector3<float>& pos = m2pg.GetSourcePositions().at(d).at(r);
            Write(ss, static_cast<float>(pos.GetX()));
            Write(ss, static_cast<float>(pos.GetY()));
            Write(ss, static_cast<float>(pos.GetZ()));
        }
    }

    for (size_t d = 0; d < nd; ++d)
    {
        for (size_t r = 0; r < nr; ++r)
        {
            for (size_t m = 0; m < nMic; ++m)
            {
                for (size_t f = 0; f < nFFT/2 + 1; ++f)
                {
                    const std::complex<float>& c = m2pg.GetTFs().at(d).at(r).at(m).at(f);
                    Write(ss, c.real());
                    Write(ss, c.imag());
                }
            }
        }
    }

    ss.seekp(0, std::ios::beg);
}

void MakeTestData(std::stringstream& ss, const harkio2::hark::hgtf::GTF1_0& gtf, uint32_t nMic)
{
    ss.write("GTF ", 4);
    uint32_t nSource = gtf.GetNSource();
    uint32_t nFIndex = gtf.GetNFIndex();
    Write(ss, static_cast<uint32_t>(8 + 12*nSource + 4 + 4*nFIndex + 8*nSource*nMic*nFIndex));
    Write(ss, static_cast<uint16_t>(1));
    Write(ss, static_cast<uint16_t>(0));

    Write(ss, nSource);

    for (size_t s = 0; s < nSource; ++s)
    {
        const harkio2::Vector3<float>& pos = gtf.GetSourcePositions().at(s);
        Write(ss, static_cast<float>(pos.GetX()));
        Write(ss, static_cast<float>(pos.GetY()));
        Write(ss, static_cast<float>(pos.GetZ()));
    }

    Write(ss, nFIndex);

    for (size_t i = 0; i < nFIndex; ++i)
    {
        Write(ss, static_cast<uint32_t>(gtf.GetFIndexToBinIndexTable()[i]));
    }

    for (size_t s = 0; s < nSource; ++s)
    {
        for (size_t m = 0; m < nMic; ++m)
        {
            for (size_t i = 0; i < nFIndex; ++i)
            {
                const std::complex<float>& c = gtf.GetTFs().at(s).at(m).at(i);
                Write(ss, c.real());
                Write(ss, c.imag());
            }
        }
    }

    ss.seekp(0, std::ios::beg);
}

void MakeTestData(std::stringstream& ss, const harkio2::hark::hgtf::SM1_0& sm, uint32_t nMic)
{
    ss.write("SM  ", 4);
    uint32_t nSource = sm.GetNSource();
    uint32_t nFIndex = sm.GetNFIndex();
    Write(ss, static_cast<uint32_t>(8 + 12*nSource + 4 + 4*nFIndex + 8*nSource*nMic*nFIndex));
    Write(ss, static_cast<uint16_t>(1));
    Write(ss, static_cast<uint16_t>(0));

    Write(ss, nSource);

    for (size_t s = 0; s < nSource; ++s)
    {
        const harkio2::Vector3<float>& pos = sm.GetSourcePositions().at(s);
        Write(ss, static_cast<float>(pos.GetX()));
        Write(ss, static_cast<float>(pos.GetY()));
        Write(ss, static_cast<float>(pos.GetZ()));
    }

    Write(ss, nFIndex);

    for (size_t i = 0; i < nFIndex; ++i)
    {
        Write(ss, static_cast<uint32_t>(sm.GetFIndexToBinIndexTable()[i]));
    }

    for (size_t s = 0; s < nSource; ++s)
    {
        for (size_t m = 0; m < nMic; ++m)
        {
            for (size_t i = 0; i < nFIndex; ++i)
            {
                const std::complex<float>& c = sm.GetSMs().at(s).at(m).at(i);
                Write(ss, c.real());
                Write(ss, c.imag());
            }
        }
    }

    ss.seekp(0, std::ios::beg);
}

BOOST_AUTO_TEST_CASE(fmt10_test)
{
    std::stringstream testdata;
    harkio2::hark::hgtf::Fmt1_0 fmt10;
    fmt10.SetFormatId(1); // M2PG
    fmt10.SetNMic(2);
    fmt10.SetNFFT(1024);
    fmt10.SetNImpulseResponse(512);
    fmt10.SetFS(16000);

    std::vector<harkio2::Vector3<float> > positions;
    for (size_t i = 1; i <= 2; ++i)
    {
        positions.push_back(harkio2::Vector3<float>(i * 10, i * 20, i * 30));
    }
    fmt10.SetMicPositions(positions);

    MakeTestData(testdata, fmt10);

    harkio2::hark::HarkFileBinaryReader reader(testdata);
    harkio2::hark::HarkFileBinaryReader::Helper helper(reader);

    boost::shared_ptr<harkio2::hark::hgtf::Fmt1_0> fmt10_ = helper.ReadFmt1_0();

    BOOST_CHECK_EQUAL(fmt10_->GetFormatId(), fmt10.GetFormatId());
    BOOST_CHECK_EQUAL(fmt10_->GetNMic(), fmt10.GetNMic());
    BOOST_CHECK_EQUAL(fmt10_->GetNFFT(), fmt10.GetNFFT());
    BOOST_CHECK_EQUAL(fmt10_->GetNImpulseResponse(), fmt10.GetNImpulseResponse());
    BOOST_CHECK_EQUAL(fmt10_->GetFS(), fmt10.GetFS());
    BOOST_CHECK(fmt10_->GetMicPositions() == fmt10.GetMicPositions());

    // make erroneous test data.
    testdata.str("");
    MakeTestData(testdata, fmt10);
    testdata.seekp(4, std::ios::beg);
    Write(testdata, static_cast<uint32_t>(24 + 12 * 1)); // invalid size value

    harkio2::hark::HarkFileBinaryReader reader2(testdata);
    harkio2::hark::HarkFileBinaryReader::Helper helper2(reader2);

    BOOST_CHECK_THROW(helper2.ReadFmt1_0(), harkio2::AnalyzeError);

    // make erroneous test data.
    testdata.str("");
    MakeTestData(testdata, fmt10);
    testdata.write("Fmt ", 4);

    harkio2::hark::HarkFileBinaryReader reader3(testdata);
    harkio2::hark::HarkFileBinaryReader::Helper helper3(reader3);

    BOOST_CHECK_THROW(helper3.ReadFmt1_0(), harkio2::AnalyzeError);
}

BOOST_AUTO_TEST_CASE(spec10_test)
{
    std::stringstream testdata;
    harkio2::hark::hgtf::Spec1_0 spec10;
    spec10.SetDate("2011/07/18");
    spec10.SetLocation("smart room");
    spec10.SetRobot("asimo");
    spec10.SetMicArrayType("8ch circular");
    spec10.SetCreator("nakadai");
    spec10.SetMemo("memo");

    MakeTestData(testdata, spec10);

    harkio2::hark::HarkFileBinaryReader reader(testdata);
    harkio2::hark::HarkFileBinaryReader::Helper helper(reader);

    boost::shared_ptr<harkio2::hark::hgtf::Spec1_0> spec10_ = helper.ReadSpec1_0();

    BOOST_CHECK_EQUAL(spec10_->GetDate(), spec10.GetDate());
    BOOST_CHECK_EQUAL(spec10_->GetLocation(), spec10.GetLocation());
    BOOST_CHECK_EQUAL(spec10_->GetRobot(), spec10.GetRobot());
    BOOST_CHECK_EQUAL(spec10_->GetMicArrayType(), spec10.GetMicArrayType());
    BOOST_CHECK_EQUAL(spec10_->GetCreator(), spec10.GetCreator());
    BOOST_CHECK_EQUAL(spec10_->GetMemo(), spec10.GetMemo());

    // make erroneous test data.
    testdata.str("");
    MakeTestData(testdata, spec10);
    testdata.seekp(4, std::ios::beg);
    Write(testdata, static_cast<uint32_t>(1550)); // invalid size value

    harkio2::hark::HarkFileBinaryReader reader2(testdata);
    harkio2::hark::HarkFileBinaryReader::Helper helper2(reader2);

    BOOST_CHECK_THROW(helper2.ReadSpec1_0(), harkio2::AnalyzeError);

    // make erroneous test data.
    testdata.str("");
    MakeTestData(testdata, spec10);
    testdata.write("SpEC", 4);

    harkio2::hark::HarkFileBinaryReader reader3(testdata);
    harkio2::hark::HarkFileBinaryReader::Helper helper3(reader3);

    BOOST_CHECK_THROW(helper3.ReadSpec1_0(), harkio2::AnalyzeError);
}

BOOST_AUTO_TEST_CASE(m2pg10_test)
{
    std::stringstream testdata;
    harkio2::hark::hgtf::M2PG1_0 m2pg10;

    m2pg10.SetND(2);
    m2pg10.SetNR(1);

    std::vector<std::vector<harkio2::Vector3<float> > > positions(
            2, std::vector<harkio2::Vector3<float> >(1, harkio2::Vector3<float>(1, 2, 3)));
    std::vector<std::vector<std::vector<harkio2::hark::hgtf::M2PG1_0::TFType>>> tfs(
            2, std::vector<std::vector<harkio2::hark::hgtf::M2PG1_0::TFType>>(
                1, std::vector<harkio2::hark::hgtf::M2PG1_0::TFType>(1,
                    harkio2::hark::hgtf::M2PG1_0::TFType(512/2 + 1, std::complex<float>(4, 5)))));

    m2pg10.SetSourcePositions(positions);
    m2pg10.SetTFs(tfs);

    MakeTestData(testdata, m2pg10, 1, 512);

    harkio2::hark::HarkFileBinaryReader reader(testdata);
    harkio2::hark::HarkFileBinaryReader::Helper helper(reader);

    boost::shared_ptr<harkio2::hark::hgtf::M2PG1_0> m2pg10_ = helper.ReadM2PG1_0(1, 512);

    BOOST_CHECK_EQUAL(m2pg10_->GetND(), m2pg10.GetND());
    BOOST_CHECK_EQUAL(m2pg10_->GetNR(), m2pg10.GetNR());

    BOOST_REQUIRE_EQUAL(m2pg10_->GetSourcePositions().size(), 2);
    BOOST_REQUIRE_EQUAL(m2pg10_->GetSourcePositions()[0].size(), 1);
    BOOST_CHECK(m2pg10_->GetSourcePositions()[0][0] == harkio2::Vector3<float>(1, 2, 3));

    BOOST_REQUIRE_EQUAL(m2pg10_->GetTFs().size(), 2);
    BOOST_REQUIRE_EQUAL(m2pg10_->GetTFs()[0].size(), 1);
    BOOST_REQUIRE_EQUAL(m2pg10_->GetTFs()[0][0].size(), 1);
    BOOST_REQUIRE_EQUAL(m2pg10_->GetTFs()[0][0][0].size(), 512/2 + 1);
    BOOST_CHECK_EQUAL(m2pg10_->GetTFs()[0][0][0][0], std::complex<float>(4, 5));

}

BOOST_AUTO_TEST_CASE(gtf10_test)
{
    std::stringstream testdata;
    harkio2::hark::hgtf::GTF1_0 gtf10;

    gtf10.SetNSource(2);

    std::vector<harkio2::Vector3<float>> positions(
            2, harkio2::Vector3<float>(1, 2, 3));
    gtf10.SetSourcePositions(positions);

    gtf10.SetNFIndex(129);

    std::vector<unsigned int> fIndexToBinIndexTable(129);
    for (size_t i = 0; i < 129; ++i) fIndexToBinIndexTable[i] = i + 128;
    gtf10.SetFIndexToBinIndexTable(fIndexToBinIndexTable);

    std::vector<std::vector<harkio2::hark::hgtf::GTF1_0::TFType>> tfs(
            2, std::vector<harkio2::hark::hgtf::GTF1_0::TFType>(
                1, harkio2::hark::hgtf::GTF1_0::TFType(129, std::complex<float>(10, 20))));
    gtf10.SetTFs(tfs);

    MakeTestData(testdata, gtf10, 1);

    harkio2::hark::HarkFileBinaryReader reader(testdata);
    harkio2::hark::HarkFileBinaryReader::Helper helper(reader);

    boost::shared_ptr<harkio2::hark::hgtf::GTF1_0> gtf10_ = helper.ReadGTF1_0(1);

    BOOST_CHECK_EQUAL(gtf10_->GetNSource(), gtf10.GetNSource());
    BOOST_CHECK_EQUAL(gtf10_->GetNFIndex(), gtf10.GetNFIndex());

    BOOST_REQUIRE_EQUAL(gtf10_->GetSourcePositions().size(), 2);
    BOOST_CHECK(gtf10_->GetSourcePositions()[0] == harkio2::Vector3<float>(1, 2, 3));

    BOOST_REQUIRE_EQUAL(gtf10_->GetFIndexToBinIndexTable().size(), 129);
    BOOST_CHECK_EQUAL(gtf10_->GetFIndexToBinIndexTable()[0], 128);
    BOOST_CHECK_EQUAL(gtf10_->GetFIndexToBinIndexTable()[128], 128+128);

    BOOST_REQUIRE_EQUAL(gtf10_->GetTFs().size(), 2);
    BOOST_REQUIRE_EQUAL(gtf10_->GetTFs()[0].size(), 1);
    BOOST_REQUIRE_EQUAL(gtf10_->GetTFs()[0][0].size(), 129);
    BOOST_CHECK_EQUAL(gtf10_->GetTFs()[0][0][0], std::complex<float>(10, 20));
    BOOST_CHECK_EQUAL(gtf10_->GetTFs()[0][0][128], std::complex<float>(10, 20));

}

BOOST_AUTO_TEST_CASE(sm10_test)
{
    std::stringstream testdata;
    harkio2::hark::hgtf::SM1_0 sm10;

    sm10.SetNSource(2);

    std::vector<harkio2::Vector3<float>> positions(
            2, harkio2::Vector3<float>(1, 2, 3));
    sm10.SetSourcePositions(positions);

    sm10.SetNFIndex(129);

    std::vector<unsigned int> fIndexToBinIndexTable(129);
    for (size_t i = 0; i < 129; ++i) fIndexToBinIndexTable[i] = i + 128;
    sm10.SetFIndexToBinIndexTable(fIndexToBinIndexTable);

    std::vector<harkio2::hark::hgtf::SM1_0::SMType> sms(
            2, harkio2::hark::hgtf::SM1_0::SMType(
                1, std::vector<std::complex<float>>(129, std::complex<float>(10, 20))));
    sm10.SetSMs(sms);

    MakeTestData(testdata, sm10, 1);

    harkio2::hark::HarkFileBinaryReader reader(testdata);
    harkio2::hark::HarkFileBinaryReader::Helper helper(reader);

    boost::shared_ptr<harkio2::hark::hgtf::SM1_0> sm10_ = helper.ReadSM1_0(1);

    BOOST_CHECK_EQUAL(sm10_->GetNSource(), sm10.GetNSource());
    BOOST_CHECK_EQUAL(sm10_->GetNFIndex(), sm10.GetNFIndex());

    BOOST_REQUIRE_EQUAL(sm10_->GetSourcePositions().size(), 2);
    BOOST_CHECK(sm10_->GetSourcePositions()[0] == harkio2::Vector3<float>(1, 2, 3));

    BOOST_REQUIRE_EQUAL(sm10_->GetFIndexToBinIndexTable().size(), 129);
    BOOST_CHECK_EQUAL(sm10_->GetFIndexToBinIndexTable()[0], 128+0);
    BOOST_CHECK_EQUAL(sm10_->GetFIndexToBinIndexTable()[128], 128+128);

    BOOST_REQUIRE_EQUAL(sm10_->GetSMs().size(), 2);
    BOOST_REQUIRE_EQUAL(sm10_->GetSMs()[0].size(), 1);
    BOOST_REQUIRE_EQUAL(sm10_->GetSMs()[0][0].size(), 129);
    BOOST_CHECK_EQUAL(sm10_->GetSMs()[0][0][0], std::complex<float>(10, 20));
    BOOST_CHECK_EQUAL(sm10_->GetSMs()[0][0][128], std::complex<float>(10, 20));

}

BOOST_AUTO_TEST_CASE(fmt_test)
{
    std::stringstream testdata;
    harkio2::hark::hgtf::Fmt1_0 fmt10;
    fmt10.SetFormatId(1); // M2PG
    fmt10.SetNMic(2);
    fmt10.SetNFFT(1024);
    fmt10.SetNImpulseResponse(512);
    fmt10.SetFS(16000);

    std::vector<harkio2::Vector3<float> > positions;
    for (size_t i = 1; i <= 2; ++i)
    {
        positions.push_back(harkio2::Vector3<float>(i * 10, i * 20, i * 30));
    }
    fmt10.SetMicPositions(positions);

    MakeTestData(testdata, fmt10);

    harkio2::hark::HarkFileBinaryReader reader(testdata);
    harkio2::hark::HarkFileBinaryReader::Helper helper(reader);

    boost::shared_ptr<harkio2::hark::hgtf::Fmt> fmt_ = helper.ReadFmt();
    BOOST_REQUIRE(fmt_->GetVersion() == harkio2::Version(1, 0));
    boost::shared_ptr<harkio2::hark::hgtf::Fmt1_0> fmt10_ =
        boost::dynamic_pointer_cast<harkio2::hark::hgtf::Fmt1_0>(fmt_);

    BOOST_CHECK_EQUAL(fmt10_->GetFormatId(), fmt10.GetFormatId());
    BOOST_CHECK_EQUAL(fmt10_->GetNMic(), fmt10.GetNMic());
    BOOST_CHECK_EQUAL(fmt10_->GetNFFT(), fmt10.GetNFFT());
    BOOST_CHECK_EQUAL(fmt10_->GetNImpulseResponse(), fmt10.GetNImpulseResponse());
    BOOST_CHECK_EQUAL(fmt10_->GetFS(), fmt10.GetFS());
    BOOST_CHECK(fmt10_->GetMicPositions() == fmt10.GetMicPositions());

    // make erroneous test data.
    testdata.str("");
    MakeTestData(testdata, fmt10);
    testdata.seekp(4, std::ios::beg);
    Write(testdata, static_cast<uint32_t>(24 + 12 * 1)); // invalid size value

    harkio2::hark::HarkFileBinaryReader reader2(testdata);
    harkio2::hark::HarkFileBinaryReader::Helper helper2(reader2);

    BOOST_CHECK_THROW(helper2.ReadFmt1_0(), harkio2::AnalyzeError);

    // make erroneous test data.
    testdata.str("");
    MakeTestData(testdata, fmt10);
    testdata.write("Fmt ", 4);

    harkio2::hark::HarkFileBinaryReader reader3(testdata);
    harkio2::hark::HarkFileBinaryReader::Helper helper3(reader3);

    BOOST_CHECK_THROW(helper3.ReadFmt1_0(), harkio2::AnalyzeError);
}

BOOST_AUTO_TEST_CASE(spec_test)
{
    std::stringstream testdata;
    harkio2::hark::hgtf::Spec1_0 spec10;
    spec10.SetDate("2011/07/18");
    spec10.SetLocation("smart room");
    spec10.SetRobot("asimo");
    spec10.SetMicArrayType("8ch circular");
    spec10.SetCreator("nakadai");
    spec10.SetMemo("memo");

    MakeTestData(testdata, spec10);

    harkio2::hark::HarkFileBinaryReader reader(testdata);
    harkio2::hark::HarkFileBinaryReader::Helper helper(reader);

    boost::shared_ptr<harkio2::hark::hgtf::Spec> spec_ = helper.ReadSpec();
    BOOST_REQUIRE(spec_->GetVersion() == harkio2::Version(1, 0));
    boost::shared_ptr<harkio2::hark::hgtf::Spec1_0> spec10_ =
        boost::dynamic_pointer_cast<harkio2::hark::hgtf::Spec1_0>(spec_);

    BOOST_CHECK_EQUAL(spec10_->GetDate(), spec10.GetDate());
    BOOST_CHECK_EQUAL(spec10_->GetLocation(), spec10.GetLocation());
    BOOST_CHECK_EQUAL(spec10_->GetRobot(), spec10.GetRobot());
    BOOST_CHECK_EQUAL(spec10_->GetMicArrayType(), spec10.GetMicArrayType());
    BOOST_CHECK_EQUAL(spec10_->GetCreator(), spec10.GetCreator());
    BOOST_CHECK_EQUAL(spec10_->GetMemo(), spec10.GetMemo());

    // make erroneous test data.
    testdata.str("");
    MakeTestData(testdata, spec10);
    testdata.seekp(4, std::ios::beg);
    Write(testdata, static_cast<uint32_t>(1550)); // invalid size value

    harkio2::hark::HarkFileBinaryReader reader2(testdata);
    harkio2::hark::HarkFileBinaryReader::Helper helper2(reader2);

    BOOST_CHECK_THROW(helper2.ReadSpec1_0(), harkio2::AnalyzeError);

    // make erroneous test data.
    testdata.str("");
    MakeTestData(testdata, spec10);
    testdata.write("SpEC", 4);

    harkio2::hark::HarkFileBinaryReader reader3(testdata);
    harkio2::hark::HarkFileBinaryReader::Helper helper3(reader3);

    BOOST_CHECK_THROW(helper3.ReadSpec1_0(), harkio2::AnalyzeError);
}

BOOST_AUTO_TEST_CASE(m2pg_test)
{
    std::stringstream testdata;
    harkio2::hark::hgtf::M2PG1_0 m2pg10;

    m2pg10.SetND(2);
    m2pg10.SetNR(1);

    std::vector<std::vector<harkio2::Vector3<float> > > positions(
            2, std::vector<harkio2::Vector3<float> >(1, harkio2::Vector3<float>(1, 2, 3)));
    std::vector<std::vector<std::vector<harkio2::hark::hgtf::M2PG1_0::TFType>>> tfs(
            2, std::vector<std::vector<harkio2::hark::hgtf::M2PG1_0::TFType>>(
                1, std::vector<harkio2::hark::hgtf::M2PG1_0::TFType>(1,
                    harkio2::hark::hgtf::M2PG1_0::TFType(512/2 + 1, std::complex<float>(4, 5)))));

    m2pg10.SetSourcePositions(positions);
    m2pg10.SetTFs(tfs);

    MakeTestData(testdata, m2pg10, 1, 512);

    harkio2::hark::HarkFileBinaryReader reader(testdata);
    harkio2::hark::HarkFileBinaryReader::Helper helper(reader);

    boost::shared_ptr<harkio2::hark::hgtf::M2PG> m2pg_ = helper.ReadM2PG(1, 512);
    BOOST_REQUIRE(m2pg_->GetVersion() == harkio2::Version(1, 0));
    boost::shared_ptr<harkio2::hark::hgtf::M2PG1_0> m2pg10_ =
        boost::dynamic_pointer_cast<harkio2::hark::hgtf::M2PG1_0>(m2pg_);

    BOOST_CHECK_EQUAL(m2pg10_->GetND(), m2pg10.GetND());
    BOOST_CHECK_EQUAL(m2pg10_->GetNR(), m2pg10.GetNR());

    BOOST_REQUIRE_EQUAL(m2pg10_->GetSourcePositions().size(), 2);
    BOOST_REQUIRE_EQUAL(m2pg10_->GetSourcePositions()[0].size(), 1);
    BOOST_CHECK(m2pg10_->GetSourcePositions()[0][0] == harkio2::Vector3<float>(1, 2, 3));

    BOOST_REQUIRE_EQUAL(m2pg10_->GetTFs().size(), 2);
    BOOST_REQUIRE_EQUAL(m2pg10_->GetTFs()[0].size(), 1);
    BOOST_REQUIRE_EQUAL(m2pg10_->GetTFs()[0][0].size(), 1);
    BOOST_REQUIRE_EQUAL(m2pg10_->GetTFs()[0][0][0].size(), 512/2 + 1);
    BOOST_CHECK_EQUAL(m2pg10_->GetTFs()[0][0][0][0], std::complex<float>(4, 5));

}

BOOST_AUTO_TEST_CASE(gtf_test)
{
    std::stringstream testdata;
    harkio2::hark::hgtf::GTF1_0 gtf10;

    gtf10.SetNSource(2);

    std::vector<harkio2::Vector3<float>> positions(
            2, harkio2::Vector3<float>(1, 2, 3));
    gtf10.SetSourcePositions(positions);

    gtf10.SetNFIndex(129);

    std::vector<unsigned int> fIndexToBinIndexTable(129);
    for (size_t i = 0; i < 129; ++i) fIndexToBinIndexTable[i] = i + 128;
    gtf10.SetFIndexToBinIndexTable(fIndexToBinIndexTable);

    std::vector<std::vector<harkio2::hark::hgtf::GTF1_0::TFType>> tfs(
            2, std::vector<harkio2::hark::hgtf::GTF1_0::TFType>(
                1, harkio2::hark::hgtf::GTF1_0::TFType(129, std::complex<float>(10, 20))));
    gtf10.SetTFs(tfs);

    MakeTestData(testdata, gtf10, 1);

    harkio2::hark::HarkFileBinaryReader reader(testdata);
    harkio2::hark::HarkFileBinaryReader::Helper helper(reader);

    boost::shared_ptr<harkio2::hark::hgtf::GTF> gtf_ = helper.ReadGTF(1);
    BOOST_REQUIRE(gtf_->GetVersion() == harkio2::Version(1, 0));
    boost::shared_ptr<harkio2::hark::hgtf::GTF1_0> gtf10_ =
        boost::dynamic_pointer_cast<harkio2::hark::hgtf::GTF1_0>(gtf_);

    BOOST_CHECK_EQUAL(gtf10_->GetNSource(), gtf10.GetNSource());
    BOOST_CHECK_EQUAL(gtf10_->GetNFIndex(), gtf10.GetNFIndex());

    BOOST_REQUIRE_EQUAL(gtf10_->GetSourcePositions().size(), 2);
    BOOST_CHECK(gtf10_->GetSourcePositions()[0] == harkio2::Vector3<float>(1, 2, 3));

    BOOST_REQUIRE_EQUAL(gtf10_->GetFIndexToBinIndexTable().size(), 129);
    BOOST_CHECK_EQUAL(gtf10_->GetFIndexToBinIndexTable()[0], 128);
    BOOST_CHECK_EQUAL(gtf10_->GetFIndexToBinIndexTable()[128], 128+128);

    BOOST_REQUIRE_EQUAL(gtf10_->GetTFs().size(), 2);
    BOOST_REQUIRE_EQUAL(gtf10_->GetTFs()[0].size(), 1);
    BOOST_REQUIRE_EQUAL(gtf10_->GetTFs()[0][0].size(), 129);
    BOOST_CHECK_EQUAL(gtf10_->GetTFs()[0][0][0], std::complex<float>(10, 20));
    BOOST_CHECK_EQUAL(gtf10_->GetTFs()[0][0][128], std::complex<float>(10, 20));

}

BOOST_AUTO_TEST_CASE(sm_test)
{
    std::stringstream testdata;
    harkio2::hark::hgtf::SM1_0 sm10;

    sm10.SetNSource(2);

    std::vector<harkio2::Vector3<float>> positions(
            2, harkio2::Vector3<float>(1, 2, 3));
    sm10.SetSourcePositions(positions);

    sm10.SetNFIndex(129);

    std::vector<unsigned int> fIndexToBinIndexTable(129);
    for (size_t i = 0; i < 129; ++i) fIndexToBinIndexTable[i] = i + 128;
    sm10.SetFIndexToBinIndexTable(fIndexToBinIndexTable);

    std::vector<harkio2::hark::hgtf::SM1_0::SMType> sms(
            2, harkio2::hark::hgtf::SM1_0::SMType(
                1, std::vector<std::complex<float>>(129, std::complex<float>(10, 20))));
    sm10.SetSMs(sms);

    MakeTestData(testdata, sm10, 1);

    harkio2::hark::HarkFileBinaryReader reader(testdata);
    harkio2::hark::HarkFileBinaryReader::Helper helper(reader);

    boost::shared_ptr<harkio2::hark::hgtf::SM> sm_ = helper.ReadSM(1);
    BOOST_REQUIRE(sm_->GetVersion() == harkio2::Version(1, 0));
    boost::shared_ptr<harkio2::hark::hgtf::SM1_0> sm10_ =
        boost::dynamic_pointer_cast<harkio2::hark::hgtf::SM1_0>(sm_);

    BOOST_CHECK_EQUAL(sm10_->GetNSource(), sm10.GetNSource());
    BOOST_CHECK_EQUAL(sm10_->GetNFIndex(), sm10.GetNFIndex());

    BOOST_REQUIRE_EQUAL(sm10_->GetSourcePositions().size(), 2);
    BOOST_CHECK(sm10_->GetSourcePositions()[0] == harkio2::Vector3<float>(1, 2, 3));

    BOOST_REQUIRE_EQUAL(sm10_->GetFIndexToBinIndexTable().size(), 129);
    BOOST_CHECK_EQUAL(sm10_->GetFIndexToBinIndexTable()[0], 128+0);
    BOOST_CHECK_EQUAL(sm10_->GetFIndexToBinIndexTable()[128], 128+128);

    BOOST_REQUIRE_EQUAL(sm10_->GetSMs().size(), 2);
    BOOST_REQUIRE_EQUAL(sm10_->GetSMs()[0].size(), 1);
    BOOST_REQUIRE_EQUAL(sm10_->GetSMs()[0][0].size(), 129);
    BOOST_CHECK_EQUAL(sm10_->GetSMs()[0][0][0], std::complex<float>(10, 20));
    BOOST_CHECK_EQUAL(sm10_->GetSMs()[0][0][128], std::complex<float>(10, 20));

}

BOOST_AUTO_TEST_CASE(hgtf_test)
{
    std::stringstream testdata;

    unsigned int nMic = 2, nFFT = 8, nd = 2, nr = 1;
    size_t fmtSize = 8 + 24 + 12 * nMic;
    size_t m2pgSize = 8 + 12 + 12 * nd * nr + 8 * nd * nr * nMic * (nFFT/2 + 1);
    size_t specSize = 8 + 1551;

    testdata.write("HARK", 4);
    Write(testdata, static_cast<uint32_t>(
                4 + fmtSize + m2pgSize + specSize
                ));
    testdata.write("HGTF", 4);

    // fmt
    harkio2::hark::hgtf::Fmt1_0 fmt10;
    fmt10.SetFormatId(1); // M2PG
    fmt10.SetNMic(nMic);
    fmt10.SetNFFT(nFFT);
    fmt10.SetNImpulseResponse(512);
    fmt10.SetFS(16000);

    std::vector<harkio2::Vector3<float> > micPositions;
    for (size_t i = 1; i <= nMic; ++i)
    {
        micPositions.push_back(harkio2::Vector3<float>(i * 10, i * 20, i * 30));
    }
    fmt10.SetMicPositions(micPositions);

    MakeTestData(testdata, fmt10);

    // m2pg
    harkio2::hark::hgtf::M2PG1_0 m2pg10;

    m2pg10.SetND(nd);
    m2pg10.SetNR(nr);

    std::vector<std::vector<harkio2::Vector3<float> > > sourcePositions(
            nd, std::vector<harkio2::Vector3<float> >(nr, harkio2::Vector3<float>(1, 2, 3)));
    std::vector<std::vector<std::vector<harkio2::hark::hgtf::M2PG1_0::TFType>>> tfs(
            nd, std::vector<std::vector<harkio2::hark::hgtf::M2PG1_0::TFType>>(
                nr, std::vector<harkio2::hark::hgtf::M2PG1_0::TFType>(
                    nMic, harkio2::hark::hgtf::M2PG1_0::TFType(nFFT/2 + 1, std::complex<float>(4, 5)))));

    m2pg10.SetSourcePositions(sourcePositions);
    m2pg10.SetTFs(tfs);

    testdata.seekp(8 + 4 + fmtSize, std::ios::beg);
    MakeTestData(testdata, m2pg10, nMic, nFFT);

    // spec
    harkio2::hark::hgtf::Spec1_0 spec10;
    spec10.SetDate("2011/07/18");
    spec10.SetLocation("smart room");
    spec10.SetRobot("asimo");
    spec10.SetMicArrayType("8ch circular");
    spec10.SetCreator("nakadai");
    spec10.SetMemo("memo");

    testdata.seekp(8 + 4 + fmtSize + m2pgSize, std::ios::beg);
    MakeTestData(testdata, spec10);

    testdata.seekg(8, std::ios::beg); // seek to "HGTF" tag
    harkio2::hark::HarkFileBinaryReader reader(testdata);
    harkio2::hark::HarkFileBinaryReader::Helper helper(reader);

    auto harkFile = helper.ReadHGTF();

    BOOST_REQUIRE_EQUAL(harkFile->GetDataSetsSize(), 3);

    auto read_fmt = harkFile->GetDataSet("hgtf::Fmt");
    auto read_fmt10 = boost::dynamic_pointer_cast<harkio2::hark::hgtf::Fmt1_0>(read_fmt);
    BOOST_REQUIRE(read_fmt10);
    BOOST_CHECK_EQUAL(read_fmt10->GetFormatId(), 1);
    BOOST_CHECK_EQUAL(read_fmt10->GetNMic(), nMic);
    BOOST_CHECK_EQUAL(read_fmt10->GetNFFT(), nFFT);
    BOOST_CHECK_EQUAL(read_fmt10->GetNImpulseResponse(), 512);
    BOOST_CHECK_EQUAL(read_fmt10->GetFS(), 16000);
    BOOST_REQUIRE_EQUAL(read_fmt10->GetMicPositions().size(), nMic);
    BOOST_CHECK(read_fmt10->GetMicPositions()[nMic - 1] == harkio2::Vector3<float>(20, 40, 60));

    auto read_m2pg = harkFile->GetDataSet("hgtf::M2PG");
    auto read_m2pg10 = boost::dynamic_pointer_cast<harkio2::hark::hgtf::M2PG1_0>(read_m2pg);
    BOOST_REQUIRE(read_m2pg10);
    BOOST_CHECK_EQUAL(read_m2pg10->GetND(), nd);
    BOOST_CHECK_EQUAL(read_m2pg10->GetNR(), nr);
    BOOST_REQUIRE_EQUAL(read_m2pg10->GetSourcePositions().size(), nd);
    BOOST_REQUIRE_EQUAL(read_m2pg10->GetSourcePositions()[nd-1].size(), nr);
    BOOST_CHECK(read_m2pg10->GetSourcePositions()[nd-1][nr-1] == harkio2::Vector3<float>(1, 2, 3));
    BOOST_REQUIRE_EQUAL(read_m2pg10->GetTFs().size(), nd);
    BOOST_REQUIRE_EQUAL(read_m2pg10->GetTFs()[nd-1].size(), nr);
    BOOST_REQUIRE_EQUAL(read_m2pg10->GetTFs()[nd-1][nr-1].size(), nMic);
    BOOST_REQUIRE_EQUAL(read_m2pg10->GetTFs()[nd-1][nr-1][nMic-1].size(), nFFT/2 + 1);
    BOOST_CHECK_EQUAL(read_m2pg10->GetTFs()[nd-1][nr-1][nMic-1][nFFT/2], std::complex<float>(4, 5));

    auto read_spec = harkFile->GetDataSet("hgtf::Spec");
    auto read_spec10 = boost::dynamic_pointer_cast<harkio2::hark::hgtf::Spec1_0>(read_spec);
    BOOST_REQUIRE(read_spec10);
    BOOST_CHECK_EQUAL(read_spec10->GetDate(), "2011/07/18");
    BOOST_CHECK_EQUAL(read_spec10->GetMemo(), "memo");
}

